@@ -1,72 +1,34 @@
-/* $Id: ParentNode.java,v 1.25 2001/02/12 19:21:19 lehors Exp $ */
 /*
- * The Apache Software License, Version 1.1
- *
- *
- * Copyright (c) 1999-2000 The Apache Software Foundation.  All rights 
- * reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer. 
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. The end-user documentation included with the redistribution,
- *    if any, must include the following acknowledgment:  
- *       "This product includes software developed by the
- *        Apache Software Foundation (http://www.apache.org/)."
- *    Alternately, this acknowledgment may appear in the software itself,
- *    if and wherever such third-party acknowledgments normally appear.
- *
- * 4. The names "Xerces" and "Apache Software Foundation" must
- *    not be used to endorse or promote products derived from this
- *    software without prior written permission. For written 
- *    permission, please contact apache@apache.org.
- *
- * 5. Products derived from this software may not be called "Apache",
- *    nor may "Apache" appear in their name, without prior written
- *    permission of the Apache Software Foundation.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
- * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
- * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
- * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Apache Software Foundation and was
- * originally based on software copyright (c) 1999, International
- * Business Machines, Inc., http://www.apache.org.  For more
- * information on the Apache Software Foundation, please see
- * <http://www.apache.org/>.
+ * Copyright 1999-2002,2004,2005 The Apache Software Foundation.
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ * 
+ *      http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
 
 package org.apache.xerces.dom;
 
-import java.io.*;
+import java.io.Serializable;
+import java.io.IOException;
+import java.io.ObjectInputStream;
+import java.io.ObjectOutputStream;
 
-import org.w3c.dom.*;
-import org.w3c.dom.events.*;
-import org.apache.xerces.dom.*;
-import org.apache.xerces.dom.events.*;
+import org.w3c.dom.DOMException;
+import org.w3c.dom.Document;
+import org.w3c.dom.Node;
+import org.w3c.dom.NodeList;
+import org.w3c.dom.UserDataHandler;
 
 /**
- * ParentNode inherits from ChildImpl and adds the capability of having child
+ * ParentNode inherits from ChildNode and adds the capability of having child
  * nodes. Not every node in the DOM can have children, so only nodes that can
  * should inherit from this class and pay the price for it.
  * <P>
@@ -86,17 +48,23 @@ import org.apache.xerces.dom.events.*;
  * called ChildAndParentNode that inherited from ChildNode. But due to the lack
  * of multiple inheritance a lot of code had to be duplicated which led to a
  * maintenance nightmare. At the same time only a few nodes (Document,
- * DocumentFragment, Entity, and Attribute) cannot be a child so the gain is
+ * DocumentFragment, Entity, and Attribute) cannot be a child so the gain in
  * memory wasn't really worth it. The only type for which this would be the
  * case is Attribute, but we deal with there in another special way, so this is
  * not applicable.
+ * <p>
+ * This class doesn't directly support mutation events, however, it notifies
+ * the document when mutations are performed so that the document class do so.
  *
  * <p><b>WARNING</b>: Some of the code here is partially duplicated in
  * AttrImpl, be careful to keep these two classes in sync!
+ * 
+ * @xerces.internal
  *
  * @author Arnaud  Le Hors, IBM
  * @author Joe Kesselman, IBM
  * @author Andy Clark, IBM
+ * @version $Id: ParentNode.java,v 1.50 2005/05/02 22:02:22 mrglavas Exp $
  */
 public abstract class ParentNode
     extends ChildNode {
@@ -105,21 +73,15 @@ public abstract class ParentNode
     static final long serialVersionUID = 2815829867152120872L;
 
     /** Owner document. */
-    protected DocumentImpl ownerDocument;
+    protected CoreDocumentImpl ownerDocument;
 
     /** First child. */
     protected ChildNode firstChild = null;
 
     // transients
 
-    /** Cached node list length. */
-    protected transient int fCachedLength = -1;
-
-    /** Last requested node. */
-    protected transient ChildNode fCachedChild;
-
-    /** Last requested node index. */
-    protected transient int fCachedChildIndex = -1;
+    /** NodeList cache */
+    protected transient NodeListCache fNodeListCache = null;
 
     //
     // Constructors
@@ -129,7 +91,7 @@ public abstract class ParentNode
      * No public constructor; only subclasses of ParentNode should be
      * instantiated, and those normally via a Document's factory methods
      */
-    protected ParentNode(DocumentImpl ownerDocument) {
+    protected ParentNode(CoreDocumentImpl ownerDocument) {
         super(ownerDocument);
         this.ownerDocument = ownerDocument;
     }
@@ -173,8 +135,7 @@ public abstract class ParentNode
     	newnode.firstChild      = null;
 
         // invalidate cache for children NodeList
-        newnode.fCachedChildIndex = -1;
-        newnode.fCachedLength = -1;
+        newnode.fNodeListCache = null;
 
         // Then, if deep, clone the kids too.
     	if (deep) {
@@ -200,9 +161,9 @@ public abstract class ParentNode
 
     /**
      * same as above but returns internal type and this one is not overridden
-     * by DocumentImpl to return null
+     * by CoreDocumentImpl to return null
      */
-    DocumentImpl ownerDocument() {
+    CoreDocumentImpl ownerDocument() {
         return ownerDocument;
     }
 
@@ -210,7 +171,7 @@ public abstract class ParentNode
      * NON-DOM
      * set the ownerDocument of this node and its children
      */
-    void setOwnerDocument(DocumentImpl doc) {
+    void setOwnerDocument(CoreDocumentImpl doc) {
         if (needsSyncChildren()) {
             synchronizeChildren();
         }
@@ -299,8 +260,8 @@ public abstract class ParentNode
      * immediately before. If refChild is null, the insertion occurs
      * after all existing Nodes, like appendChild().
      *
-     * @returns newChild, in its new state (relocated, or emptied in the
-     * case of DocumentNode.)
+     * @return newChild, in its new state (relocated, or emptied in the case of
+     * DocumentNode.)
      *
      * @throws DOMException(HIERARCHY_REQUEST_ERR) if newChild is of a
      * type that shouldn't be a child of this node, or if newChild is an
@@ -318,7 +279,7 @@ public abstract class ParentNode
     public Node insertBefore(Node newChild, Node refChild) 
         throws DOMException {
         // Tail-call; optimizer should be able to do good things with.
-        return internalInsertBefore(newChild,refChild,MUTATION_ALL);
+        return internalInsertBefore(newChild, refChild, false);
     } // insertBefore(Node,Node):Node
      
     /** NON-DOM INTERNAL: Within DOM actions,we sometimes need to be able
@@ -326,7 +287,7 @@ public abstract class ParentNode
      * insertBefore operation allows us to do so. It is not intended
      * for use by application programs.
      */
-    Node internalInsertBefore(Node newChild, Node refChild,int mutationMask) 
+    Node internalInsertBefore(Node newChild, Node refChild, boolean replace) 
         throws DOMException {
 
         boolean errorChecking = ownerDocument.errorChecking;
@@ -355,8 +316,8 @@ public abstract class ParentNode
 
                     if (!ownerDocument.isKidOK(this, kid)) {
                         throw new DOMException(
-                                           DOMException.HIERARCHY_REQUEST_ERR, 
-                                           "DOM006 Hierarchy request error");
+                              DOMException.HIERARCHY_REQUEST_ERR, 
+                              DOMMessageFormatter.formatMessage(DOMMessageFormatter.DOM_DOMAIN, "HIERARCHY_REQUEST_ERR", null));
                     }
                 }
             }
@@ -382,21 +343,21 @@ public abstract class ParentNode
         if (errorChecking) {
             if (isReadOnly()) {
                 throw new DOMException(
-                                     DOMException.NO_MODIFICATION_ALLOWED_ERR, 
-                                       "DOM001 Modification not allowed");
+                              DOMException.NO_MODIFICATION_ALLOWED_ERR, 
+                              DOMMessageFormatter.formatMessage(DOMMessageFormatter.DOM_DOMAIN, "NO_MODIFICATION_ALLOWED_ERR", null));
             }
-            if (newChild.getOwnerDocument() != ownerDocument) {
+            if (newChild.getOwnerDocument() != ownerDocument && newChild != ownerDocument) {
                 throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, 
-                                       "DOM005 Wrong document");
+                            DOMMessageFormatter.formatMessage(DOMMessageFormatter.DOM_DOMAIN, "WRONG_DOCUMENT_ERR", null));
             }
             if (!ownerDocument.isKidOK(this, newChild)) {
                 throw new DOMException(DOMException.HIERARCHY_REQUEST_ERR, 
-                                       "DOM006 Hierarchy request error");
+                            DOMMessageFormatter.formatMessage(DOMMessageFormatter.DOM_DOMAIN, "HIERARCHY_REQUEST_ERR", null));
             }
             // refChild must be a child of this node (or null)
             if (refChild != null && refChild.getParentNode() != this) {
                 throw new DOMException(DOMException.NOT_FOUND_ERR,
-                                       "DOM008 Not found");
+                            DOMMessageFormatter.formatMessage(DOMMessageFormatter.DOM_DOMAIN, "NOT_FOUND_ERR", null));
             }
 
             // Prevent cycles in the tree
@@ -409,22 +370,12 @@ public abstract class ParentNode
             }
             if(!treeSafe) {
                 throw new DOMException(DOMException.HIERARCHY_REQUEST_ERR, 
-                                       "DOM006 Hierarchy request error");
+                            DOMMessageFormatter.formatMessage(DOMMessageFormatter.DOM_DOMAIN, "HIERARCHY_REQUEST_ERR", null));
             }
         }
 
-        EnclosingAttr enclosingAttr=null;
-        if (MUTATIONEVENTS && ownerDocument.mutationEvents
-            && (mutationMask&MUTATION_AGGREGATE)!=0) {
-            // MUTATION PREPROCESSING
-            // No direct pre-events, but if we're within the scope 
-            // of an Attr and DOMAttrModified was requested,
-            // we need to preserve its previous value.
-            LCount lc=LCount.lookup(MutationEventImpl.DOM_ATTR_MODIFIED);
-            if (lc.captures+lc.bubbles+lc.defaults>0) {
-                enclosingAttr=getEnclosingAttr();
-            }
-        }
+        // notify document
+        ownerDocument.insertingNode(this, replace);
 
         // Convert to internal type, to avoid repeated casting
         ChildNode newInternal = (ChildNode)newChild;
@@ -482,80 +433,30 @@ public abstract class ParentNode
         changed();
 
         // update cached length if we have any
-        if (fCachedLength != -1) {
-            fCachedLength++;
-        }
-        if (fCachedChildIndex != -1) {
-            // if we happen to insert just before the cached node, update
-            // the cache to the new node to match the cached index
-            if (fCachedChild == refInternal) {
-                fCachedChild = newInternal;
-            } else {
-                // otherwise just invalidate the cache
-                fCachedChildIndex = -1;
+        if (fNodeListCache != null) {
+            if (fNodeListCache.fLength != -1) {
+                fNodeListCache.fLength++;
             }
-        }
-
-        if (MUTATIONEVENTS && ownerDocument.mutationEvents) {
-            // MUTATION POST-EVENTS:
-            // "Local" events (non-aggregated)
-            if ((mutationMask&MUTATION_LOCAL) != 0) {
-                // New child is told it was inserted, and where
-                LCount lc = LCount.lookup(MutationEventImpl.DOM_NODE_INSERTED);
-                if (lc.captures+lc.bubbles+lc.defaults>0) {
-                    MutationEvent me= new MutationEventImpl();
-                    me.initMutationEvent(MutationEventImpl.DOM_NODE_INSERTED,
-                                         true,false,this,null,
-                                          null,null,(short)0);
-                    newInternal.dispatchEvent(me);
-                }
-
-                // If within the Document, tell the subtree it's been added
-                // to the Doc.
-                lc=LCount.lookup(
-                            MutationEventImpl.DOM_NODE_INSERTED_INTO_DOCUMENT);
-                if (lc.captures+lc.bubbles+lc.defaults>0) {
-                    NodeImpl eventAncestor=this;
-                    if (enclosingAttr!=null) 
-                        eventAncestor=
-                            (NodeImpl)(enclosingAttr.node.getOwnerElement());
-                    if (eventAncestor!=null) { // Might have been orphan Attr
-                        NodeImpl p=eventAncestor;
-                        while (p!=null) {
-                            eventAncestor=p; // Last non-null ancestor
-                            // In this context, ancestry includes
-                            // walking back from Attr to Element
-                            if(p.getNodeType()==ATTRIBUTE_NODE) {
-                                p=(ElementImpl)((AttrImpl)p).getOwnerElement();
-                            }
-                            else {
-                                p=p.parentNode();
-                            }
-                        }
-                        if(eventAncestor.getNodeType()==Node.DOCUMENT_NODE) {
-                            MutationEvent me= new MutationEventImpl();
-                            me.initMutationEvent(MutationEventImpl
-                                              .DOM_NODE_INSERTED_INTO_DOCUMENT,
-                                                 false,false,null,null,
-                                                 null,null,(short)0);
-                            dispatchEventToSubtree(newInternal,me);
-                        }
-                    }
+            if (fNodeListCache.fChildIndex != -1) {
+                // if we happen to insert just before the cached node, update
+                // the cache to the new node to match the cached index
+                if (fNodeListCache.fChild == refInternal) {
+                    fNodeListCache.fChild = newInternal;
+                } else {
+                    // otherwise just invalidate the cache
+                    fNodeListCache.fChildIndex = -1;
                 }
             }
-
-            // Subroutine: Transmit DOMAttrModified and DOMSubtreeModified
-            // (Common to most kinds of mutation)
-            if ((mutationMask&MUTATION_AGGREGATE) != 0) {
-                dispatchAggregateEvents(enclosingAttr);
-            }
         }
 
+        // notify document
+        ownerDocument.insertedNode(this, newInternal, replace);
+
         checkNormalizationAfterInsert(newInternal);
 
         return newChild;
 
-    } // internalInsertBefore(Node,Node,int):Node
+    } // internalInsertBefore(Node,Node,boolean):Node
 
     /**
      * Remove a child from this Node. The removed child's subtree
@@ -572,7 +473,7 @@ public abstract class ParentNode
     public Node removeChild(Node oldChild) 
         throws DOMException {
         // Tail-call, should be optimizable
-        return internalRemoveChild(oldChild,MUTATION_ALL);
+        return internalRemoveChild(oldChild, false);
     } // removeChild(Node) :Node
      
     /** NON-DOM INTERNAL: Within DOM actions,we sometimes need to be able
@@ -580,98 +481,42 @@ public abstract class ParentNode
      * removeChild operation allows us to do so. It is not intended
      * for use by application programs.
      */
-    Node internalRemoveChild(Node oldChild,int mutationMask)
+    Node internalRemoveChild(Node oldChild, boolean replace)
         throws DOMException {
 
-        DocumentImpl ownerDocument = ownerDocument();
+        CoreDocumentImpl ownerDocument = ownerDocument();
         if (ownerDocument.errorChecking) {
             if (isReadOnly()) {
                 throw new DOMException(
-                                     DOMException.NO_MODIFICATION_ALLOWED_ERR, 
-                                     "DOM001 Modification not allowed");
+                            DOMException.NO_MODIFICATION_ALLOWED_ERR, 
+                            DOMMessageFormatter.formatMessage(DOMMessageFormatter.DOM_DOMAIN, "NO_MODIFICATION_ALLOWED_ERR", null));
             }
             if (oldChild != null && oldChild.getParentNode() != this) {
                 throw new DOMException(DOMException.NOT_FOUND_ERR, 
-                                       "DOM008 Not found");
+                            DOMMessageFormatter.formatMessage(DOMMessageFormatter.DOM_DOMAIN, "NOT_FOUND_ERR", null));
             }
         }
 
-        // notify document
-        ownerDocument.removedChildNode(oldChild);
-
         ChildNode oldInternal = (ChildNode) oldChild;
 
-        EnclosingAttr enclosingAttr=null;
-        if(MUTATIONEVENTS && ownerDocument.mutationEvents)
-        {
-            // MUTATION PREPROCESSING AND PRE-EVENTS:
-            // If we're within the scope of an Attr and DOMAttrModified 
-            // was requested, we need to preserve its previous value for
-            // that event.
-            LCount lc=LCount.lookup(MutationEventImpl.DOM_ATTR_MODIFIED);
-            if(lc.captures+lc.bubbles+lc.defaults>0)
-            {
-                enclosingAttr=getEnclosingAttr();
-            }
-            
-            if( (mutationMask&MUTATION_LOCAL) != 0)
-            {
-                // Child is told that it is about to be removed
-                lc=LCount.lookup(MutationEventImpl.DOM_NODE_REMOVED);
-                if(lc.captures+lc.bubbles+lc.defaults>0)
-                {
-                    MutationEvent me= new MutationEventImpl();
-                    me.initMutationEvent(MutationEventImpl.DOM_NODE_REMOVED,
-                                         true,false,this,null,
-                                         null,null,(short)0);
-                    oldInternal.dispatchEvent(me);
-                }
-            
-                // If within Document, child's subtree is informed that it's
-                // losing that status
-                lc=LCount.lookup(
-                             MutationEventImpl.DOM_NODE_REMOVED_FROM_DOCUMENT);
-                if(lc.captures+lc.bubbles+lc.defaults>0)
-                {
-                    NodeImpl eventAncestor=this;
-                    if(enclosingAttr!=null) 
-                        eventAncestor=
-                            (NodeImpl) enclosingAttr.node.getOwnerElement();
-                    if(eventAncestor!=null) // Might have been orphan Attr
-                    {
-                        for(NodeImpl p=eventAncestor.parentNode();
-                            p!=null;
-                            p=p.parentNode())
-                        {
-                            eventAncestor=p; // Last non-null ancestor
-                        }
-                        if(eventAncestor.getNodeType()==Node.DOCUMENT_NODE)
-                        {
-                            MutationEvent me= new MutationEventImpl();
-                            me.initMutationEvent(MutationEventImpl
-                                               .DOM_NODE_REMOVED_FROM_DOCUMENT,
-                                                 false,false,
-                                                 null,null,null,null,(short)0);
-                            dispatchEventToSubtree(oldInternal,me);
-                        }
-                    }
-                }
-            }
-        } // End mutation preprocessing
+        // notify document
+        ownerDocument.removingNode(this, oldInternal, replace);
 
         // update cached length if we have any
-        if (fCachedLength != -1) {
-            fCachedLength--;
-        }
-        if (fCachedChildIndex != -1) {
-            // if the removed node is the cached node
-            // move the cache to its (soon former) previous sibling
-            if (fCachedChild == oldInternal) {
-                fCachedChildIndex--;
-                fCachedChild = oldInternal.previousSibling();
-            } else {
-                // otherwise just invalidate the cache
-                fCachedChildIndex = -1;
+        if (fNodeListCache != null) {
+            if (fNodeListCache.fLength != -1) {
+                fNodeListCache.fLength--;
+            }
+            if (fNodeListCache.fChildIndex != -1) {
+                // if the removed node is the cached node
+                // move the cache to its (soon former) previous sibling
+                if (fNodeListCache.fChild == oldInternal) {
+                    fNodeListCache.fChildIndex--;
+                    fNodeListCache.fChild = oldInternal.previousSibling();
+                } else {
+                    // otherwise just invalidate the cache
+                    fNodeListCache.fChildIndex = -1;
+                }
             }
         }
 
@@ -709,20 +554,14 @@ public abstract class ParentNode
 
         changed();
 
-        if(MUTATIONEVENTS && ownerDocument.mutationEvents)
-        {
-            // MUTATION POST-EVENTS:
-            // Subroutine: Transmit DOMAttrModified and DOMSubtreeModified,
-            // if required. (Common to most kinds of mutation)
-            if( (mutationMask&MUTATION_AGGREGATE) != 0)
-                dispatchAggregateEvents(enclosingAttr);
-        } // End mutation postprocessing
+        // notify document
+        ownerDocument.removedNode(this, replace);
 
         checkNormalizationAfterRemove(oldPreviousSibling);
 
         return oldInternal;
 
-    } // internalRemoveChild(Node,int):Node
+    } // internalRemoveChild(Node,boolean):Node
 
     /**
      * Make newChild occupy the location that oldChild used to
@@ -730,7 +569,7 @@ public abstract class ParentNode
      * parent, if any. Equivalent to inserting newChild before oldChild,
      * then removing oldChild.
      *
-     * @returns oldChild, in its new state (removed).
+     * @return oldChild, in its new state (removed).
      *
      * @throws DOMException(HIERARCHY_REQUEST_ERR) if newChild is of a
      * type that shouldn't be a child of this node, or if newChild is
@@ -753,33 +592,79 @@ public abstract class ParentNode
         // this as either desirable or undesirable, but hints that
         // aggregations should be issued only once per user request.
 
-        EnclosingAttr enclosingAttr=null;
-        if(MUTATIONEVENTS && ownerDocument.mutationEvents)
-        {
-            // MUTATION PREPROCESSING AND PRE-EVENTS:
-            // If we're within the scope of an Attr and DOMAttrModified 
-            // was requested, we need to preserve its previous value for
-            // that event.
-            LCount lc=LCount.lookup(MutationEventImpl.DOM_ATTR_MODIFIED);
-            if(lc.captures+lc.bubbles+lc.defaults>0)
-            {
-                enclosingAttr=getEnclosingAttr();
-            }
-        } // End mutation preprocessing
+        // notify document
+        ownerDocument.replacingNode(this);
 
-        internalInsertBefore(newChild, oldChild,MUTATION_LOCAL);
+        internalInsertBefore(newChild, oldChild, true);
         if (newChild != oldChild) {
-            internalRemoveChild(oldChild,MUTATION_LOCAL);
+            internalRemoveChild(oldChild, true);
         }
 
-        if(MUTATIONEVENTS && ownerDocument.mutationEvents)
-        {
-            dispatchAggregateEvents(enclosingAttr);
-        }
+        // notify document
+        ownerDocument.replacedNode(this);
 
         return oldChild;
     }
 
+    /*
+     * Get Node text content
+     * @since DOM Level 3
+     */
+    public String getTextContent() throws DOMException {
+        Node child = getFirstChild();
+        if (child != null) {
+            Node next = child.getNextSibling();
+            if (next == null) {
+                return hasTextContent(child) ? ((NodeImpl) child).getTextContent() : "";
+            }
+            if (fBufferStr == null){
+                fBufferStr = new StringBuffer();
+            }
+            else {
+                fBufferStr.setLength(0);
+            }
+            getTextContent(fBufferStr);
+            return fBufferStr.toString();
+        }
+        return "";
+    }
+
+    // internal method taking a StringBuffer in parameter
+    void getTextContent(StringBuffer buf) throws DOMException {
+        Node child = getFirstChild();
+        while (child != null) {
+            if (hasTextContent(child)) {
+                ((NodeImpl) child).getTextContent(buf);
+            }
+            child = child.getNextSibling();
+        }
+    }
+
+    // internal method returning whether to take the given node's text content
+    final boolean hasTextContent(Node child) {
+        return child.getNodeType() != Node.COMMENT_NODE &&
+            child.getNodeType() != Node.PROCESSING_INSTRUCTION_NODE &&
+            (child.getNodeType() != Node.TEXT_NODE ||
+             ((TextImpl) child).isIgnorableWhitespace() == false);
+    }
+
+    /*
+     * Set Node text content
+     * @since DOM Level 3
+     */
+    public void setTextContent(String textContent)
+        throws DOMException {
+        // get rid of any existing children
+        Node child;
+        while ((child = getFirstChild()) != null) {
+            removeChild(child);
+        }
+        // create a Text node to hold the given content
+        if (textContent != null && textContent.length() != 0){
+            appendChild(ownerDocument().createTextNode(textContent));
+        }
+    }
+
     //
     // NodeList methods
     //
@@ -791,22 +676,37 @@ public abstract class ParentNode
      */
     private int nodeListGetLength() {
 
-        if (fCachedLength == -1) { // is the cached length invalid ?
-            ChildNode node;
+        if (fNodeListCache == null) {
+            // get rid of trivial cases
+            if (firstChild == null) {
+                return 0;
+            }
+            if (firstChild == lastChild()) {
+                return 1;
+            }
+            // otherwise request a cache object
+            fNodeListCache = ownerDocument.getNodeListCache(this);
+        }
+        if (fNodeListCache.fLength == -1) { // is the cached length invalid ?
+            int l;
+            ChildNode n;
             // start from the cached node if we have one
-            if (fCachedChildIndex != -1 && fCachedChild != null) {
-                fCachedLength = fCachedChildIndex;
-                node = fCachedChild;
+            if (fNodeListCache.fChildIndex != -1 &&
+                fNodeListCache.fChild != null) {
+                l = fNodeListCache.fChildIndex;
+                n = fNodeListCache.fChild;
             } else {
-                node = firstChild;
-                fCachedLength = 0;
+                n = firstChild;
+                l = 0;
             }
-            for (; node != null; node = node.nextSibling) {
-                fCachedLength++;
+            while (n != null) {
+                l++;
+                n = n.nextSibling;
             }
+            fNodeListCache.fLength = l;
         }
 
-        return fCachedLength;
+        return fNodeListCache.fLength;
 
     } // nodeListGetLength():int
 
@@ -824,31 +724,61 @@ public abstract class ParentNode
      * @param index int
      */
     private Node nodeListItem(int index) {
+
+        if (fNodeListCache == null) {
+            // get rid of trivial case
+            if (firstChild == lastChild()) {
+                return index == 0 ? firstChild : null;
+            }
+            // otherwise request a cache object
+            fNodeListCache = ownerDocument.getNodeListCache(this);
+        }
+        int i = fNodeListCache.fChildIndex;
+        ChildNode n = fNodeListCache.fChild;
+        boolean firstAccess = true;
         // short way
-        if (fCachedChildIndex != -1 && fCachedChild != null) {
-            if (fCachedChildIndex < index) {
-                while (fCachedChildIndex < index && fCachedChild != null) {
-                    fCachedChildIndex++;
-                    fCachedChild = fCachedChild.nextSibling;
+        if (i != -1 && n != null) {
+            firstAccess = false;
+            if (i < index) {
+                while (i < index && n != null) {
+                    i++;
+                    n = n.nextSibling;
                 }
             }
-            else if (fCachedChildIndex > index) {
-                while (fCachedChildIndex > index && fCachedChild != null) {
-                    fCachedChildIndex--;
-                    fCachedChild = fCachedChild.previousSibling();
+            else if (i > index) {
+                while (i > index && n != null) {
+                    i--;
+                    n = n.previousSibling();
                 }
             }
-            return fCachedChild;
+        }
+        else {
+            // long way
+            if (index < 0) {
+                return null;
+            }
+            n = firstChild;
+            for (i = 0; i < index && n != null; i++) {
+                n = n.nextSibling;
+            }
         }
 
-        // long way
-        fCachedChild = firstChild;
-        for (fCachedChildIndex = 0; 
-             fCachedChildIndex < index && fCachedChild != null; 
-             fCachedChildIndex++) {
-            fCachedChild = fCachedChild.nextSibling;
+        // release cache if reaching last child or first child
+        if (!firstAccess && (n == firstChild || n == lastChild())) {
+            fNodeListCache.fChildIndex = -1;
+            fNodeListCache.fChild = null;
+            ownerDocument.freeNodeListCache(fNodeListCache);
+            // we can keep using the cache until it is actually reused
+            // fNodeListCache will be nulled by the pool (document) if that
+            // happens.
+            // fNodeListCache = null;
         }
-        return fCachedChild;
+        else {
+            // otherwise update it
+            fNodeListCache.fChildIndex = i;
+            fNodeListCache.fChild = n;
+        }
+        return n;
 
     } // nodeListItem(int):Node
 
@@ -919,6 +849,32 @@ public abstract class ParentNode
         isNormalized(true);
     }
 
+    /**
+     * DOM Level 3 WD- Experimental.
+     * Override inherited behavior from NodeImpl to support deep equal.
+     */
+    public boolean isEqualNode(Node arg) {
+        if (!super.isEqualNode(arg)) {
+            return false;
+        }
+        // there are many ways to do this test, and there isn't any way
+        // better than another. Performance may vary greatly depending on
+        // the implementations involved. This one should work fine for us.
+        Node child1 = getFirstChild();
+        Node child2 = arg.getFirstChild();
+        while (child1 != null && child2 != null) {
+            if (!((NodeImpl) child1).isEqualNode(child2)) {
+                return false;
+            }
+            child1 = child1.getNextSibling();
+            child2 = child2.getNextSibling();
+        }
+        if (child1 != child2) {
+            return false;
+        }
+        return true;
+    }
+
     //
     // Public methods
     //
@@ -965,55 +921,6 @@ public abstract class ParentNode
         needsSyncChildren(false);
     }
 
-    /**
-     * Synchronizes the node's children with the internal structure.
-     * Fluffing the children at once solves a lot of work to keep
-     * the two structures in sync. The problem gets worse when
-     * editing the tree -- this makes it a lot easier.
-     * Even though this is only used in deferred classes this method is
-     * put here so that it can be shared by all deferred classes.
-     */
-    protected final void synchronizeChildren(int nodeIndex) {
-
-        // we don't want to generate any event for this so turn them off
-        boolean orig = ownerDocument.mutationEvents;
-        ownerDocument.mutationEvents = false;
-
-        // no need to sync in the future
-        needsSyncChildren(false);
-
-        // create children and link them as siblings
-        DeferredDocumentImpl ownerDocument =
-            (DeferredDocumentImpl)this.ownerDocument;
-        ChildNode first = null;
-        ChildNode last = null;
-        for (int index = ownerDocument.getLastChild(nodeIndex);
-             index != -1;
-             index = ownerDocument.getPrevSibling(index)) {
-
-            ChildNode node = (ChildNode)ownerDocument.getNodeObject(index);
-            if (last == null) {
-                last = node;
-            }
-            else {
-                first.previousSibling = node;
-            }
-            node.ownerNode = this;
-            node.isOwned(true);
-            node.nextSibling = first;
-            first = node;
-        }
-        if (last != null) {
-            firstChild = first;
-            first.isFirstChild(true);
-            lastChild(last);
-        }
-
-        // set mutation events flag back to its original value
-        ownerDocument.mutationEvents = orig;
-
-    } // synchronizeChildren()
-
     /**
      * Checks the normalized state of this node after inserting a child.
      * If the inserted child causes this node to be unnormalized, then this
@@ -1099,15 +1006,24 @@ public abstract class ParentNode
         // perform default deseralization
         ois.defaultReadObject();
 
-        // hardset synchildren - so we don't try to sync- it does not make any sense
-        // to try to synchildren when we just desealize object.
-
+        // hardset synchildren - so we don't try to sync - it does not make any
+        // sense to try to synchildren when we just deserialize object.
         needsSyncChildren(false);
 
-        // initialize transients
-        fCachedLength = -1;
-        fCachedChildIndex = -1;
-
     } // readObject(ObjectInputStream)
 
+    /*
+     * a class to store some user data along with its handler
+     */
+    class UserDataRecord implements Serializable {
+        /** Serialization version. */
+        private static final long serialVersionUID = 3258126977134310455L;
+        
+        Object fData;
+        UserDataHandler fHandler;
+        UserDataRecord(Object data, UserDataHandler handler) {
+            fData = data;
+            fHandler = handler;
+        }
+    }
 } // class ParentNode
